// for neatnessi the dio.h header is shared with user space 
// so the includes go here:  

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/moduleparam.h>

#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/delay.h>

#include <linux/config.h>
#include <linux/pci.h>  

#include "dio_interface.h"
#include "dio.h"


#define DIO_MOD_NAME	"dio"

MODULE_LICENSE("GPL");

#define DIO_MAJOR 0
static int dio_major=DIO_MAJOR;
module_param(dio_major, int, 0);


/*MODULE_AUTHOR("Matt Clark, Aug 2001");*/

static int	dio_ioctl(struct inode *, struct file *, unsigned int , unsigned long);
static int	dio_open(struct inode *, struct file *);
static int	dio_release(struct inode *, struct file *);
static int	init_dio(void);
static void	cleanup_dio(void);
static ssize_t	dio_read(struct file * , char *, size_t, loff_t *);
static ssize_t	dio_write(struct file *, const char *, size_t, loff_t *);
inline static void dio_delay(void);


module_init(init_dio);
module_exit(cleanup_dio);

struct file_operations dio_fops = {
	ioctl:		dio_ioctl,
	open:		dio_open,
	release:	dio_release,
	read:		dio_read,
	write:		dio_write,
	};

int init_dio(void) {
int res;
volatile u32	temp;
struct pci_dev*	dev=NULL;

	printk("dio dev: init_module: Matt Clark Aug 2005- device %d\n",dio_major);
/* look for a DIO card */
	if(!(dev=pci_find_device(DIO_VENDOR,DIO_DEVICE,dev))){
		printk("dio: no dio pci card found in this machine\n");
		return -ENODEV;
		}
/* next error should <never> occur (here for coding errors only)*/
	if(dev==NULL){
		printk("dio: card found but dev==NULL\n");
		return -ENODEV;
		}
/* now we have the device and a pointer to its device structure
 * extract the numbers we need using the dev structure rather than
 * the configuration registers (change since 2.1.xx matt oct 2001)
 * in case the kernel has rewritten the locations 
 */
/* now I know that base_address[0] is config
 *		 base_address[1] is the FPGA fuse map and private device code
 *		 base_address[2] is the I/O ports (as pci memory)  
 * I will assume that the base address is static once I get it 
 */
	dio_base=ioremap(pci_resource_start(dev,2)&PCI_BASE_ADDRESS_MEM_MASK,DIO_IO_REGION_SIZE);
	if(dio_base==0){
		printk("dio: init: could not ioremap pci address space\n");
		return -ENODEV;
		}
	dio_cntl=(unsigned int *)((void *)dio_base+DIO_CNTL);
	PRINT_DEBUG("dio: init: DIO at %x and CNTL at %x\n",(int) dio_base,(int) dio_cntl);
	

/* now configure the control word */
	temp=readl(dio_cntl);
	PRINT_DEBUG("dio:init:cntl=%x\n",temp);
	writel(0,dio_cntl);
	dio_dir=0;
	temp=readl(dio_cntl);
	PRINT_DEBUG("dio:init:cntl=%x\n",temp);
	writel((dio_dir|dio_latch)&DIO_CNTL_PORT_MASK,dio_cntl);
	temp=readl(dio_cntl);
	PRINT_DEBUG("dio:init:cntl=%x\n",temp);

/* now register the char device - never sure what order is best */
	res=register_chrdev(dio_major,DIO_MOD_NAME,&dio_fops);
	if (res<0) return(res);
	dio_major=res;
	printk("dio dev: init_module: registered on %d\n",(int)res);
	
	return 0;
	}     

/* OPEN AND RELEASE */
static int dio_open(struct inode *inode, struct file *file){ 
	if((file->private_data=kmalloc(sizeof(u32)*DIO_BUFF_SIZE,GFP_KERNEL))==NULL)
		return -ENOMEM;
	return 0;
	}

static int dio_release(struct inode *inode, struct file *file) { 
	kfree(file->private_data);
	return 0;
	}	

static ssize_t dio_read(struct file * file, char * buf, size_t count, loff_t *ppos){
// this is a test, just read port0
u32 *buff;
int length,i,ccount;
	if((dio_read_seq&(DIO_PORT_0|DIO_PORT_1|DIO_PORT_2))==0){
		printk("Failed on the read_seq\n");
		return -EINVAL;
		}
	if(dio_read_n==0){
		printk("Failed on the dio_read_n value\n");
		return -EINVAL;
		}
	buff=file->private_data;
	// clip and quantise count
	ccount=count;
	if(count>DIO_BUFF_SIZE)ccount=DIO_BUFF_SIZE;
	length=dio_read_n*(ccount/(sizeof(u32)*dio_read_n));
	if(length==0){
		printk("Failed on the length\n");
		return -EINVAL;	
		}
	for(i=0;i<length;){
		if(dio_read_seq&DIO_PORT_0)
			buff[i++]=readl(dio_base+DIO_IO_PORT_0);
		if(dio_read_seq&DIO_DELAY_0)dio_delay();
		if(dio_read_seq&DIO_PORT_1)
			buff[i++]=readl(dio_base+DIO_IO_PORT_1);
		if(dio_read_seq&DIO_DELAY_1)dio_delay();
		if(dio_read_seq&DIO_PORT_2)
			buff[i++]=readl(dio_base+DIO_IO_PORT_2);
		if(dio_read_seq&DIO_DELAY_2)dio_delay();
		}
	length=i*sizeof(u32);
	copy_to_user(buf,buff,length);
	return length;
	}

static ssize_t dio_write(struct file * file, const char * buf, size_t count, loff_t *ppos){
// this is a test, just read port0
u32 *buff;
int length,i;
	if((dio_write_seq&(DIO_PORT_0|DIO_PORT_1|DIO_PORT_2))==0)
		return -EINVAL;
	if(dio_write_n==0)
		return -EINVAL;
	buff=file->private_data;
	// clip and quantise count
	if(count>DIO_BUFF_SIZE)count=DIO_BUFF_SIZE;
	length=dio_write_n*(count/(sizeof(u32)*dio_write_n));
	if(length==0)
		return -EINVAL;	
	copy_from_user(buff,buf,length*sizeof(u32));
	for(i=0;i<length;){
		if(dio_write_seq&DIO_PORT_0)writel(buff[i++],dio_base+DIO_IO_PORT_0);
		if(dio_write_seq&DIO_DELAY_0)dio_delay();
		if(dio_write_seq&DIO_PORT_1)writel(buff[i++],dio_base+DIO_IO_PORT_1);
		if(dio_write_seq&DIO_DELAY_1)dio_delay();
		if(dio_write_seq&DIO_PORT_2)writel(buff[i++],dio_base+DIO_IO_PORT_2);
		if(dio_write_seq&DIO_DELAY_2)dio_delay();
		}
	length=i*sizeof(u32);
	return length;
	}

static int dio_ioctl(struct inode * inode,struct file *file, unsigned int cmd, unsigned long arg){
	switch(cmd){
		case DIO_READ_SEQ:
			dio_read_seq=arg;
			dio_read_n=0;
			if(dio_read_seq&DIO_PORT_0)dio_read_n++;
			if(dio_read_seq&DIO_PORT_1)dio_read_n++;
			if(dio_read_seq&DIO_PORT_2)dio_read_n++;
			return dio_read_n;
			break;
		case DIO_WRITE_SEQ:
			dio_write_seq=arg;
			dio_write_n=0;
			if(dio_write_seq&DIO_PORT_0)dio_write_n++;
			if(dio_write_seq&DIO_PORT_1)dio_write_n++;
			if(dio_write_seq&DIO_PORT_2)dio_write_n++;
			return dio_write_n;
			break;
		case DIO_WRITE_0:
			writel((u32)arg,dio_base+DIO_IO_PORT_0);	
			break;
		case DIO_WRITE_1:
			writel((u32)arg,dio_base+DIO_IO_PORT_1);	
			break;
		case DIO_WRITE_2:
			writel((u32)arg,dio_base+DIO_IO_PORT_2);	
			break;
		case DIO_READ_0:
			return put_user(readl(dio_base+DIO_IO_PORT_0),(u32 __user *)arg);
			break;
		case DIO_READ_1:
			return put_user(readl(dio_base+DIO_IO_PORT_1),(u32 __user *)arg);
			break;
		case DIO_READ_2:
			return put_user(readl(dio_base+DIO_IO_PORT_2),(u32 __user *)arg);
			break;
		case DIO_LATCH:
			dio_latch=DIO_CNTL_LATCH_MASK;
			writel((dio_dir|dio_latch)&DIO_CNTL_PORT_MASK,dio_cntl);
			break;
		case DIO_UNLATCH:
			dio_latch=0;
			writel((dio_dir|dio_latch)&DIO_CNTL_PORT_MASK,dio_cntl);
			break;
		case DIO_SET_DIR:
			dio_dir=arg;
			writel((dio_dir|dio_latch)&DIO_CNTL_PORT_MASK,dio_cntl);
			break;
		case DIO_GET_DIR:
			put_user(dio_dir,(u32 __user *)arg);
			break;
		default:
			PRINT_DEBUG("dio: ioctl: Unknown ioctl cmd %x\n",(int)cmd);
			return -ENOTTY;
			}
	return 0;
	}  

void cleanup_dio(void){
int res;
	printk("dio dev: cleanup_module: Matt Clark Aug 2005- device %d\n",dio_major);
	if(dio_base!=NULL){
		iounmap(dio_base);
		dio_base=NULL;
		dio_cntl=NULL;
		}
	res=unregister_chrdev(dio_major,DIO_MOD_NAME);
	}    

inline static void dio_delay(void){
int jif;
        if(dio_delay_time){
                if(dio_delay_time<DIO_USECS_PER_HZ){
                /* busy wait for ~ the right time*/
                        udelay(dio_delay_time);
                        }
                else{
                /* sleep for ~ the right time */
                        jif=(dio_delay_time*HZ)/1000000;
                        if(jif<1)jif=1;
                        current->state=TASK_INTERRUPTIBLE;
                        schedule_timeout(jif);
                        }
                }
        }
